# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from .emoji import Emoji
from ..utils.api_object import APIObject


@dataclass
class Reaction(APIObject):
    """
    Represents a Discord Reaction object

    :param count:
        times this emoji has been used to react

    :param me:
        whether the current user reacted using this emoji

    :param emoji:
        emoji information
    """
    count: int
    me: bool
    emoji: Emoji



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import Callable, Any, Optional, TYPE_CHECKING
from .types import T, MISSING

if TYPE_CHECKING:
    from pincer.client import Client


def convert(
        value: Any,
        factory: Callable[[Any], T],
        check: Optional[type] = None,
        client: Optional[Client] = None
) -> T:
    """
    Convert a value to T if its not MISSING.
    """
    def handle_factory() -> T:
        def fin_fac(v: Any):
            if check is not None and isinstance(v, check):
                return v

            if client is None:
                return factory(v)

            return factory(v | {"_client": client, "_http": client.http})

        return (
            list(map(fin_fac, value))
            if isinstance(value, list)
            else fin_fac(value)
        )

    return MISSING if value is MISSING else handle_factory()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import List

from .emoji import Emoji
from ..utils import APIObject, APINullable, MISSING


@dataclass
class SelectOption(APIObject):
    """
    Represents a Discord Select Option object

    :param label:
        the user-facing name of the option, max 100 characters

    :param value:
        the def-defined value of the option, max 100 characters

    :param description:
        an additional description of the option, max 100 characters

    :param emoji:
        `id`, `name`, and `animated`

    :param default:
        will render this option as selected by default
    """
    label: str
    value: str
    description: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    default: APINullable[bool] = MISSING


@dataclass
class SelectMenu(APIObject):
    """
    Represents a Discord Select Menu object

    :param type:
        `3` for a select menu

    :param custom_id:
        a developer-defined identifier for the button,
        max 100 characters

    :param options:
        the choices in the select, max 25

    :param placeholder:
        custom placeholder text if nothing is selected,
        max 100 characters

    :param min_values:
        the minimum number of items that must be chosen;
        default 1, min 0, max 25

    :param max_values:
        the maximum number of items that can be chosen;
         default 1, max 25

    :param disabled:
        disable the select, default False
    """
    type: int
    custom_id: str
    options: List[SelectOption]

    placeholder: APINullable[str] = MISSING
    min_values: APINullable[int] = 1
    max_values: APINullable[int] = 1
    disabled: APINullable[bool] = False



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake


@dataclass
class VoiceServerUpdateEvent(APIObject):
    """
    Sent when a guild's voice server is updated.
    This is sent when initially connecting to voice,
    and when the current voice instance fails over to a new server.

    :param token:
        voice connection token

    :param guild_id:
        the guild this voice server update is for

    :param endpoint:
        the voice server host
    """
    token: str
    guild_id: Snowflake
    endpoint: Optional[str] = None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.exceptions import PincerError
from pincer.utils import APIObject


@dataclass
class DiscordError(PincerError, APIObject):
    """
    Represents an error event in the Discord Gateway.

    :param code:
        The RPC error code.

    :param message:
        The error description..
    """
    code: int
    message: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ..utils import APIObject, Snowflake


@dataclass
class FollowedChannel(APIObject):
    """Represents a Discord Followed Channel object

    :param channel_id:
        source channel id

    :param webhook_id:
        created target webhook id
    """
    channel_id: Snowflake
    webhook_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List, TYPE_CHECKING

from pincer.core.http import HTTPClient

from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp, \
    convert

if TYPE_CHECKING:
    from .. import Client


@dataclass
class GuildMember(APIObject):
    """
    Represents a member which resides in a guild/server.

    :param _client:
        reference to the Client

    :param _http:
        reference to the HTTPClient

    :param deaf:
        whether the user is deafened in voice channels

    :param joined_at:
        when the user joined the guild

    :param mute:
        whether the user is muted in voice channels

    :param roles:
        array of role object ids

    :param nick:
        this users guild nickname

    :param pending:
        whether the user has not yet passed the guild's Membership
        Screening requirements

    :param is_pending:
        Deprecated version of pending.

    :param permissions:
        total permissions of the member in the channel,
        including overwrites, returned when in the interaction object

    :param premium_since:
        when the user started boosting the guild

    :param user:
        the user this guild member represents
    """

    _client: Client
    _http: HTTPClient

    deaf: bool
    joined_at: Timestamp
    mute: bool
    roles: List[Snowflake]

    hoisted_role: APINullable[Snowflake] = MISSING
    nick: APINullable[Optional[str]] = MISSING
    pending: APINullable[bool] = MISSING
    is_pending: APINullable[bool] = MISSING
    permissions: APINullable[str] = MISSING
    premium_since: APINullable[Optional[Timestamp]] = MISSING
    user: APINullable[User] = MISSING
    avatar: APINullable[str] = MISSING

    def __post_init__(self):
        self.roles = convert(self.roles, Snowflake.from_string)
        self.user = convert(self.user, User.from_dict, User, self._client)
        self.premium_since = convert(
            self.premium_since, Timestamp.parse, Timestamp
        )

    @classmethod
    async def from_id(
            cls,
            client: Client,
            guild_id: int,
            _id: int
    ) -> GuildMember:
        data = await client.http.get(f"guilds/{guild_id}/members/{_id}")
        return cls.from_dict(data | {"_client": client, "_http": client.http})



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from ..core.dispatch import GatewayDispatch


async def payload_middleware(self, payload: GatewayDispatch):
    """Invoked when basically anything is received from gateway.."""
    return "on_payload", [payload]


def export():
    """Export the middleware"""
    return payload_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.types import MISSING, APINullable


@dataclass
class IntegrationDeleteEvent(APIObject):
    """
    Sent when an integration is deleted.

    :param id:
        integration id

    :param guild_id:
        id of the guild

    :param application_id:
        id of the bot/OAuth2 application for this discord integration
    """
    id: Snowflake
    guild_id: Snowflake
    application_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import List

from pincer.objects.channel import Channel
from pincer.objects.thread import ThreadMember
from pincer.utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class ThreadListSyncEvent(APIObject):
    """
    Sent when the current user gains access to a channel.

    :param guild_id:
        the id of the guild

    :param channel_ids:
        the parent channel ids whose threads are being synced.
        If omitted, then threads were synced for the entire guild.
        This array may contain channel_ids that have no active
        threads as well, so you know to clear that data.

    :param threads:
        all active threads in the given channels that
        the current user can access

    :param members:
        all thread member objects from the synced threads for
        the current user, indicating which threads the current
        user has been added to
    """
    guild_id: Snowflake
    threads: List[Channel]
    members: List[ThreadMember]

    channel_ids: APINullable[List[Snowflake]] = MISSING


@dataclass
class ThreadMembersUpdateEvent(APIObject):
    """
    Sent when anyone is added to or removed from a thread.
    If the current user does not have the `GUILD_MEMBERS`
    Gateway Intent, then this event will only be sent if
    the current user was added to or removed from the thread.

    :param id:
        the id of the thread

    :param guild_id:
        the id of the guild

    :param member_count:
        the approximate number of members in the thread, capped at 50

    :param added_members:
        the users who were added to the thread

    :param removed_member_ids:
        the id of the users who were removed from the thread
    """
    id: Snowflake
    guild_id: Snowflake
    member_count: int

    added_members: APINullable[List[ThreadMember]] = MISSING
    removed_member_ids: APINullable[List[Snowflake]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when the client receives a notification
(mention or new message in eligible channels)
"""

# TODO: Implement event



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import TypeVar, Callable, Coroutine, Any, Union


class MissingType:
    def __repr__(self):
        return "<MISSING>"


MISSING = MissingType()

T = TypeVar('T')

# Represents a value which is optionally returned from the API
APINullable = Union[T, MissingType]

# Represents a coroutine.
Coro = TypeVar("Coro", bound=Callable[..., Coroutine[Any, Any, Any]])



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ..utils.api_object import APIObject


@dataclass
class SessionStartLimit(APIObject):
    """
    Represents a Discord Session Start Limit object

    :param total:
        The total number of session starts the current user is allowed

    :param remaining:
        The remaining number of session starts
        the current user is allowed

    :param reset_after:
        The number of milliseconds after which the limit resets

    :param max_concurrency:
        The number of identify requests allowed per 5 seconds
    """
    total: int
    remaining: int
    reset_after: int
    max_concurrency: int



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
Sent when the user clicks a Rich Presence join invite in chat
to join a game.
"""

# TODO: Implement event




# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Dict, Any, Optional

from .context import Context
from ..utils.slidingwindow import SlidingWindow


class ThrottleScope(Enum):
    """
    On what the cooldown should be set/on what should the cooldown be
    set.

    :param GUILD:
        The cooldown is per guild.

    :param CHANNEL:
        The cooldown is per channel.

    :param USER:
        The cooldown is per user.

    :param GLOBAL:
        The cooldown is global.
    """
    GUILD = auto()
    CHANNEL = auto()
    USER = auto()
    GLOBAL = auto()


class ThrottleInterface(ABC):
    @abstractmethod
    def __init__(self, client, **kwargs):
        """
        Initializes a throttler.

        :param client:
            The current connected client, aka your bot.

        :param kwargs:
            Optional extra arguments.
        """
        self.client = client

    @abstractmethod
    async def handle(self, ctx: Context, **kwargs):
        """
        Handles a context. This method is executed before the command is.

        :param ctx:
            The context of the command.

        :param kwargs:
            The extra kwargs passed for the cooldown.
        """
        raise NotImplementedError


class DefaultThrottleHandler(ThrottleInterface, ABC):
    default_throttle_type = ThrottleScope.GLOBAL
    throttle: Dict[Any, SlidingWindow] = {}

    __throttle_scopes = {
        ThrottleScope.GLOBAL: None,
        ThrottleScope.GUILD: "guild.id",
        ThrottleScope.CHANNEL: "channel.id",
        ThrottleScope.USER: "author.id"
    }

    def get_key_from_scope(
            self,
            ctx: Context,
            scope: ThrottleScope
    ) -> Optional[int]:
        """
        Retrieve the the appropriate key from the context through the
        throttle scope.
        """
        ...

    async def handle(self, ctx: Context, **kwargs):
        ...



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when a user's voice state changes in a subscribed voice channel
(mute, volume, etc.)
"""

# TODO: Implement event



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum, Enum
from typing import List, Optional, Union, TYPE_CHECKING


from .application import Application
from .attachment import Attachment
from .channel import Channel, ChannelMention
from .embed import Embed
from .guild_member import GuildMember
from .interaction_base import MessageInteraction
from .message_component import MessageComponent
from .message_reference import MessageReference
from .reaction import Reaction
from .role import Role
from .sticker import StickerItem
from .user import User
from ..core.http import HTTPClient
from .._config import GatewayConfig
from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp, \
    convert

if TYPE_CHECKING:
    from .. import Client


class MessageActivityType(IntEnum):
    """
    The activity people can perform on a rich presence activity.

    Such an activity could for example be a spotify listen.
    """
    JOIN = 1
    SPECTATE = 2
    LISTEN = 3
    JOIN_REQUEST = 5


class MessageFlags(IntEnum):
    """
    Special message properties.

    :param CROSSPOSTED:
        the message has been published to subscribed
        channels (via Channel Following)

    :param IS_CROSSPOST:
        this message originated from a message
        in another channel (via Channel Following)

    :param SUPPRESS_EMBEDS:
        do not include any embeds when serializing this message

    :param SOURCE_MESSAGE_DELETED:
        the source message for this crosspost
        has been deleted (via Channel Following)

    :param URGENT:
        this message came from the urgent message system

    :param HAS_THREAD:
        this message has an associated thread,
        with the same id as the message

    :param EPHEMERAL:
        this message is only visible to the user
        who invoked the Interaction

    :param LOADING:
        this message is an Interaction
        Response and the bot is "thinking"
    """
    CROSSPOSTED = 1 << 0
    IS_CROSSPOST = 1 << 1
    SUPPRESS_EMBEDS = 1 << 2
    SOURCE_MESSAGE_DELETED = 1 << 3
    URGENT = 1 << 4
    HAS_THREAD = 1 << 5
    EPHEMERAL = 1 << 6
    LOADING = 1 << 7


class MessageType(IntEnum):
    """
    Represents the type of the message.
    """
    DEFAULT = 0
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    CHANNEL_NAME_CHANGE = 4
    CHANNEL_ICON_CHANGE = 5
    CHANNEL_PINNED_MESSAGE = 6
    GUILD_MEMBER_JOIN = 7
    USER_PREMIUM_GUILD_SUBSCRIPTION = 8
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1 = 9
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2 = 10
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    APPLICATION_COMMAND = 20

    if GatewayConfig.version < 8:
        REPLY = 0
        APPLICATION_COMMAND = 0

    if GatewayConfig.version >= 9:
        THREAD_STARTER_MESSAGE = 21

    GUILD_INVITE_REMINDER = 22


@dataclass
class MessageActivity(APIObject):
    """
    Represents a Discord Message Activity object

    :param type:
        type of message activity

    :param party_id:
        party_id from a Rich Presence event
    """
    type: MessageActivityType
    party_id: APINullable[str] = MISSING


class AllowedMentionTypes(str, Enum):
    """
    The allowed mentions.

    :param ROLES:
        Controls role mentions

    :param USERS:
        Controls user mentions

    :param EVERYONE:
        Controls @everyone and @here mentions
    """
    ROLES = "roles"
    USERS = "user"
    EVERYONE = "everyone"


@dataclass
class UserMessage(APIObject):
    """
    Represents a message sent in a channel within Discord.

    :param _client:

    :param _http:

    :param id:
        id of the message

    :param channel_id:
        id of the channel the message was sent in

    :param guild_id:
        id of the guild the message was sent in

    :param author:
        the author of this message (not guaranteed to be a valid user)

    :param member:
        member properties for this message's author

    :param content:
        contents of the message

    :param timestamp:
        when this message was sent

    :param edited_timestamp:
        when this message was edited (or null if never)

    :param tts:
        whether this was a TTS message

    :param mention_everyone:
        whether this message mentions everyone

    :param mentions:
        users specifically mentioned in the message

    :param mention_roles:
        roles specifically mentioned in this message

    :param mention_channels:
        channels specifically mentioned in this message

    :param attachments:
        any attached files

    :param embeds:
        any embedded content

    :param reactions:
        reactions to the message

    :param nonce:
        user for validating a message was sent

    :param pinned:
        whether this message is pinned

    :param webhook_id:
        if the message is generated by a webhook,
        this is the webhook's id

    :param type:
        type of message

    :param activity:
        sent with Rich Presence-related chat embeds

    :param application:
        sent with Rich Presence-related chat embeds

    :param application_id:
        if the message is a response to an Interaction,
        this is the id of the interaction's application

    :param message_reference:
        data showing the source of a crosspost,
        channel follow add, pin, or reply message

    :param flags:
        message flags combined as a bitfield

    :param referenced_message:
        the message associated with the message_reference

    :param interaction:
        sent if the message is a response to an Interaction

    :param thread:
        the thread that was started from this message,
        includes thread member object

    :param components:
        sent if the message contains components like buttons,
        action rows, or other interactive components

    :param sticker_items:
        sent if the message contains stickers
    """
    _client: Client
    _http: HTTPClient

    id: Snowflake
    channel_id: Snowflake
    author: User
    content: str
    timestamp: Timestamp
    edited_timestamp: Optional[Timestamp]
    tts: bool
    mention_everyone: bool
    mentions: List[User]
    mention_roles: List[Role]
    attachments: List[Attachment]
    embeds: List[Embed]
    pinned: bool
    type: MessageType

    mention_channels: APINullable[List[ChannelMention]] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    reactions: APINullable[List[Reaction]] = MISSING
    nonce: APINullable[Union[int, str]] = MISSING
    webhook_id: APINullable[Snowflake] = MISSING
    activity: APINullable[MessageActivity] = MISSING
    application: APINullable[Application] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    message_reference: APINullable[MessageReference] = MISSING
    flags: APINullable[MessageFlags] = MISSING
    referenced_message: APINullable[Optional[Message]] = MISSING
    interaction: APINullable[MessageInteraction] = MISSING
    thread: APINullable[Channel] = MISSING
    components: APINullable[List[MessageComponent]] = MISSING
    sticker_items: APINullable[List[StickerItem]] = MISSING

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)
        self.channel_id = convert(self.channel_id, Snowflake.from_string)
        self.author = convert(self.author, User.from_dict, client=self._client)
        self.timestamp = convert(self.timestamp, Timestamp)
        self.edited_timestamp = convert(self.edited_timestamp, Timestamp)
        self.mentions = convert(self.mentions, User.from_dict, client=self._client)
        self.mention_roles = convert(self.mention_roles, Role.from_dict)
        self.attachments = convert(self.attachments, Attachment.from_dict)
        self.embeds = convert(self.embeds, Embed.from_dict)
        self.mention_channels = convert(
            self.mention_channels,
            ChannelMention.from_dict
        )
        self.guild_id = convert(self.guild_id, Snowflake.from_string)
        self.member = convert(self.member, GuildMember.from_dict, client=self._client)
        self.reactions = convert(self.reactions, Reaction.from_dict)
        self.webhook_id = convert(self.webhook_id, Snowflake.from_string)
        self.activity = convert(self.activity, MessageActivity.from_dict)
        self.application = convert(self.application, Application.from_dict)
        self.application_id = convert(
            self.application_id,
            Snowflake.from_string
        )
        self.message_reference = convert(
            self.message_reference,
            MessageReference.from_dict
        )
        # self.flags = convert(self.flags, MessageFlags.from_bytes)
        # self.referenced_message = convert(
        #     self.referenced_message,
        #     Message.from_dict
        # )
        self.interaction = convert(
            self.interaction,
            MessageInteraction.from_dict
        )
        self.thread = convert(self.thread, Channel.from_dict, client=self._client)
        self.components = convert(self.components, MessageComponent.from_dict)
        self.sticker_items = convert(self.sticker_items, StickerItem.from_dict)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

import logging
from glob import glob
from importlib import import_module
from os import chdir
from pathlib import Path
from typing import Dict

from pincer.exceptions import NoExportMethod
from pincer.utils import Coro

_log = logging.getLogger(__package__)


def get_middleware() -> Dict[str, Coro]:
    middleware_list: Dict[str, Coro] = {}
    chdir(Path(__file__).parent.resolve())

    for middleware_path in glob("*.py"):
        if middleware_path.startswith("__"):
            continue

        event = middleware_path[:-3]

        try:
            middleware_list[event] = getattr(
                import_module(f".{event}", package=__name__),
                "export"
            )()
        except AttributeError:
            _log.warning(
                f"Middleware {middleware_path} expected an `export` method."
            )

            continue

            # raise NoExportMethod(
            #    f"Middleware module `{middleware_path}` expected an "
            #    "`export` method but none was found!"
            # )

    return middleware_list


middleware: Dict[str, Coro] = get_middleware()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import List, Tuple

from ...objects.application import Application
from ...objects.guild import Guild
from ...objects.user import User
from ...utils.api_object import APIObject
from ...utils.types import MISSING, APINullable


@dataclass
class HelloEvent(APIObject):
    """
    Sent on connection to the websocket.
    Defines the heartbeat interval that the client should heartbeat to.

    :param heartbeat_interval:
        the interval (in milliseconds) the client should heartbeat with
    """
    heartbeat_interval: int


@dataclass
class ReadyEvent(APIObject):
    """
    Dispatched when a client has completed the initial
    handshake with the gateway (for new sessions).

    :param v:
        gateway version

    :param user:
        information about the user including email

    :param guilds:
        the guilds the user is in

    :param session_id:
        used for resuming connections

    :param shard:
        the shard information associated
        with this session, if sent when identifying

    :param application:
        contains `id` and `flags`
    """
    v: int
    user: User
    guilds: List[Guild]
    session_id: str
    application: Application

    shard: APINullable[Tuple[int, int]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

import asyncio
import logging
from json import dumps
from typing import Dict, Any, Optional, Protocol

from aiohttp import ClientSession, ClientResponse
from aiohttp.client import _RequestContextManager
from aiohttp.typedefs import StrOrURL

from . import __package__
from .._config import GatewayConfig
from ..exceptions import (
    NotFoundError, BadRequestError, NotModifiedError, UnauthorizedError,
    ForbiddenError, MethodNotAllowedError, RateLimitError, ServerError,
    HTTPError
)

_log = logging.getLogger(__package__)


class HttpCallable(Protocol):
    """aiohttp HTTP method"""
    __name__: str

    def __call__(
            self, url: StrOrURL, *,
            allow_redirects: bool = True, json: Dict = None, **kwargs: Any
    ) -> _RequestContextManager:
        ...


class HTTPClient:
    """Interacts with Discord API through HTTP protocol"""

    def __init__(self, token: str, *, version: int = None, ttl: int = 5):
        """
        Instantiate a new HttpApi object.

        :param token:
            Discord API token

        Keyword Arguments:

        :param version:
            The discord API version.
            See `<https://discord.com/developers/docs/reference#api-versioning>`_.

        :param ttl:
            Max amount of attempts after error code 5xx
        """
        version = version or GatewayConfig.version
        self.url: str = f"https://discord.com/api/v{version}"
        self.max_ttl: int = ttl

        headers: Dict[str, str] = {
            "Authorization": f"Bot {token}",
            "Content-Type": "application/json"
        }
        self.__session: ClientSession = ClientSession(headers=headers)

        self.__http_exceptions: Dict[int, HTTPError] = {
            304: NotModifiedError(),
            400: BadRequestError(),
            401: UnauthorizedError(),
            403: ForbiddenError(),
            404: NotFoundError(),
            405: MethodNotAllowedError(),
            429: RateLimitError()
        }

    # for with block
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def close(self):
        """Closes :attr:`~.HTTPClient.__session`"""
        await self.__session.close()

    async def __send(
            self,
            method: HttpCallable,
            endpoint: str, *,
            data: Optional[Dict] = None,
            __ttl: int = None
    ) -> Optional[Dict]:
        """
        Send an api request to the Discord REST API.

        :meta public:

        :param method:
            The method for the request. (eg GET or POST)

        :param endpoint:
            The endpoint to which the request will be sent.

        Keyword Arguments:

        :param data:
            The data which will be added to the request.

        :param __ttl:
            Private param used for recursively setting the retry amount.
            (Eg set to 1 for 1 max retry)
        """
        ttl = __ttl or self.max_ttl

        if ttl == 0:
            logging.error(
                # TODO: print better method name
                f"{method.__name__.upper()} {endpoint} has reached the "
                f"maximum retry count of {self.max_ttl}."
            )

            raise ServerError(f"Maximum amount of retries for `{endpoint}`.")

        # TODO: print better method name
        _log.debug(f"{method.__name__.upper()} {endpoint} | {dumps(data)}")

        url = f"{self.url}/{endpoint}"
        async with method(url, json=data) as res:
            return await self.__handle_response(
                res, method, endpoint, data, ttl
            )

    async def __handle_response(
            self,
            res: ClientResponse,
            method: HttpCallable,
            endpoint: str,
            data: Optional[Dict],
            __ttl: int,
    ) -> Optional[Dict]:
        """
        Handle responses from the discord API.

        :meta public:

        Side effects:
            If a 5xx error code is returned it will retry the request.

        :param res:
            The response from the discord API.

        :param method:
            The method which was used to call the endpoint.

        :param endpoint:
            The endpoint to which the request was sent.

        :param data:
            The data which was added to the request.

        :param __ttl:
            Private param used for recursively setting the retry amount.
            (Eg set to 1 for 1 max retry)
        """
        _log.debug(f"Received response for {endpoint} | {await res.text()}")
        if res.ok:
            if res.status == 204:
                _log.debug(
                    "Request has been sent successfully. "
                )
                return

            _log.debug(
                "Request has been sent successfully. "
                "Returning json response."
            )

            return await res.json()

        exception = self.__http_exceptions.get(res.status)

        if exception:
            _log.error(
                f"An http exception occurred while trying to send "
                f"a request to {endpoint}. ({res.status}, {res.reason})"
            )

            exception.__init__(res.reason)
            exception.json = await res.json()
            raise exception

        # status code is guaranteed to be 5xx
        retry_in = 1 + (self.max_ttl - __ttl) * 2

        _log.debug(
            "Server side error occurred with status code "
            f"{res.status}. Retrying in {retry_in}s."
        )

        await asyncio.sleep(retry_in)

        # try sending it again
        return await self.__send(method, endpoint, __ttl=__ttl - 1, data=data)

    async def delete(self, route: str) -> Optional[Dict]:
        """
        Sends a delete request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a delete request to.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.delete, route)

    async def get(self, route: str) -> Optional[Dict]:
        """
        Sends a get request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a get request to.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.get, route)

    async def head(self, route: str) -> Optional[Dict]:
        """
        Sends a head request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a head request to.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.head, route)

    async def options(self, route: str) -> Optional[Dict]:
        """
        Sends a options request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a options request to.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.options, route)

    async def patch(self, route: str, data: Dict) -> Optional[Dict]:
        """
        Sends a patch request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a patch request to.

        :param data:
            The update data for the patch request.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.patch, route, data=data)

    async def post(self, route: str, data: Dict) -> Optional[Dict]:
        """
        Sends a post request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a post request to.

        :param data:
            The data for the post request.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.post, route, data=data)

    async def put(self, route: str, data: Dict) -> Optional[Dict]:
        """
        Sends a put request to a Discord REST endpoint.

        :param route:
            The Discord REST endpoint to send a put request to.

        :param data:
            The data for the put request.

        :return:
            JSON response from the discord API.
        """
        return await self.__send(self.__session.put, route, data=data)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from .user import User
from ..utils import APIObject


@dataclass
class Ban(APIObject):
    """
    Representation of the Discord Ban object

    :param reason:
        The reason for the ban

    :param user:
        The banned user
    """
    reason: Optional[str]
    user: User



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from .guild_member import GuildMember
from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp


@dataclass
class VoiceState(APIObject):
    """
    Used to represent a user's voice connection status

    :param guild_id:
        the guild id this voice state is for

    :param channel_id:
        the channel id this user is connected to

    :param user_id:
        the user id this voice state is for

    :param member:
        the guild member this voice state is for

    :param session_id:
        the session id for this voice state

    :param deaf:
        whether this user is deafened by the server

    :param mute:
        whether this user is muted by the server

    :param self_deaf:
        whether this user is locally deafened

    :param self_mute:
        whether this user is locally muted

    :param self_stream:
        whether this user is streaming using "Go Live"

    :param self_video:
        whether this user's camera is enabled

    :param suppress:
        whether this user is muted by the current user

    :param request_to_speak_timestamp:
        the time at which the user requested to speak
    """
    channel_id: Optional[Snowflake]
    user_id: Snowflake
    session_id: str
    deaf: bool
    mute: bool
    self_deaf: bool
    self_mute: bool
    self_video: bool
    suppress: bool
    request_to_speak_timestamp: Optional[Timestamp]

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    self_stream: APINullable[bool] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


class Snowflake(int):
    """
    Discord utilizes Twitter's snowflake format for uniquely
    identifiable descriptors (IDs).

    These IDs are guaranteed to be unique across all of Discord,
    except in some unique scenarios in which child objects
    share their parent's ID.

    Because Snowflake IDs are up to 64 bits in size (e.g. a uint64),
    they are always returned as strings in the HTTP API
    to prevent integer overflows in some languages.
    """
    @classmethod
    def from_string(cls, string: str):
        """
        Initialize a new Snowflake from a string.

        :param string:
            The snowflake as a string.
        """
        return Snowflake(int(string))

    @property
    def timestamp(self) -> int:
        """
        Milliseconds since Discord Epoch,
        the first second of 2015 or 14200704000000
        """
        return self >> 22

    @property
    def worker_id(self) -> int:
        """Internal worker ID"""
        return (self >> 17) % 16

    @property
    def process_id(self) -> int:
        """Internal process ID"""
        return (self >> 12) % 16

    @property
    def increment(self) -> int:
        """
        For every ID that is generated on that process,
        this number is incremented
        """
        return self % 2048



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import copy
from dataclasses import dataclass, fields, _is_dataclass_instance
from enum import Enum
from typing import Dict, Union, Generic, TypeVar, Any

from .types import MissingType

T = TypeVar("T")


def _asdict_ignore_none(obj: Generic[T]) -> Union[tuple, dict, T]:
    """
    Returns a dict from a dataclass that ignores
    all values that are None
    Modification of _asdict_inner from dataclasses

    :param obj:
        Dataclass obj
    """

    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _asdict_ignore_none(getattr(obj, f.name))

            if isinstance(value, Enum):
                result.append((f.name, value.value))
            # This if statement was added to the function
            elif not isinstance(value, MissingType):
                result.append((f.name, value))

        return dict(result)

    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):
        return type(obj)(*[_asdict_ignore_none(v) for v in obj])

    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_ignore_none(v) for v in obj)

    elif isinstance(obj, dict):
        return type(obj)(
            (
                _asdict_ignore_none(k),
                _asdict_ignore_none(v)
            ) for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


@dataclass
class APIObject:
    """
    Represents an object which has been fetched from the Discord API.
    """

    # def __post_init__(self):
    #     fin = {
    #         key: _eval_type(ForwardRef(value), globals(), globals())
    #         for key, value in self.__annotations__.items()
    #     }
    #
    #     # pprint(self.__annotations__)
    #     # pprint(get_type_hints(self))
    #     print(fin)
    #     print("Post init", self)

    @classmethod
    def from_dict(
            cls: Generic[T],
            data: Dict[str, Union[str, bool, int, Any]]
    ) -> T:
        """
        Parse an API object from a dictionary.
        """
        if isinstance(data, cls):
            return data

        # Disable inspection for IDE because this is valid code for the
        # inherited classes:
        # noinspection PyArgumentList
        return cls(
            **{
                key: (
                    value.value if isinstance(value, Enum) else value
                ) for key, value in data.items()
            }
        )

    def to_dict(self) -> Dict:
        """
        Transform the current object to a dictionary representation.
        """
        return _asdict_ignore_none(self)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from enum import IntEnum


class Intents(IntEnum):
    """
    Discord client intents.

    These give your client more permissions.

    NOTE: The given Intents must also be enabled for your client on
    the discord dashboard.
    """
    NONE = 0
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_BANS = 1 << 2
    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14

    @staticmethod
    def all():
        """Consists of all intents"""
        res = 0

        for intent in list(map(lambda itm: itm.value, Intents)):
            res |= intent

        return res



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from inspect import signature
from typing import Any, Optional, Protocol, TypeVar

from .insertion import should_pass_cls
from .types import Coro

T = TypeVar("T")


class GetItem(Protocol):
    """Represents a class which implements the __getitem__ property."""

    def __getitem__(self, key: int) -> Any:
        return ...


def get_index(
        collection: GetItem,
        index: int,
        fallback: Optional[T] = None
) -> Optional[T]:
    """
    Gets an item from a collection through index.
    Allows you to provide a fallback for if that index is out of bounds.

    :param collection:
        The collection from which the item is retrieved.

    :param index:
        The index of the item in the collection.

    :param fallback:
        The fallback value which will be used if the index doesn't
        exist. Default value is None.

    :return:
        The item at the provided index from the collection, or if that
        item doesn't exist it will return the fallback value.
    """
    try:
        return collection[index]

    except IndexError:
        return fallback


def get_signature_and_params(func: Coro):
    """
    Get the signature and its parameters from a coroutine.

    :param func:
        The coroutine from whom the information should be extracted.
    """
    sig = signature(func).parameters
    params = list(sig)

    if should_pass_cls(func):
        del params[0]

    return sig, params


def get_params(func: Coro):
    """
    Get the parameters from a coroutine.

    :param func:
        The coroutine from whom the information should be extracted.
    """
    _, params = get_signature_and_params(func)
    return params



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List

from ..utils import APIObject, Snowflake


@dataclass
class WelcomeScreenChannel(APIObject):
    """
    Represents a welcome screen channel. This is a channel which gets
    shown on the welcome screen.

    :param channel_id:
        the channel's id

    :param description:
        the description shown for the channel

    :param emoji_id:
        the emoji id, if the emoji is custom

    :param emoji_name:
        the emoji name if custom, the unicode character if standard,
        or null if no emoji is set
    """

    channel_id: Snowflake
    description: str

    emoji_id: Optional[int] = None
    emoji_name: Optional[str] = None


@dataclass
class WelcomeScreen(APIObject):
    """
    Representation of a Discord guild/server welcome screen.

    :description:
        the server description shown in the welcome screen

    :welcome_channels:
        the channels shown in the welcome screen, up to 5
    """
    welcome_channels: List[WelcomeScreenChannel]

    description: Optional[str] = None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ..utils import APIObject, Snowflake


@dataclass
class Overwrite(APIObject):
    """
    Represents a Discord Overwrite object

    :param id:
        role or user id

    :param type:
        either 0 (role) or 1 (member)

    :param allow:
        permission bit set

    :param deny:
        permission bit set
    """
    id: Snowflake
    type: int
    allow: str
    deny: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from enum import auto, Enum
from typing import List, Optional, Tuple, Union

from .presence import Activity
from pincer.objects.intents import Intents
from pincer.utils.api_object import APIObject
from pincer.utils.types import APINullable, MISSING
from pincer.utils.snowflake import Snowflake


@dataclass
class Identify(APIObject):
    """
    Used to trigger the initial handshake with the gateway.

    :param token:
        authentication token

    :param properties:
        connection properties

    :param compress:
        whether this connection supports compression of packets

    :param large_threshold:
        value between 50 and 250, total number
        of members where the gateway will stop sending offline
        members in the guild member list

    :param shard:
        used for Guild Sharding

    :param presence:
        presence structure for initial presence information

    :param intents:
        the Gateway Intents you wish to receive
    """
    token: str
    properties: dict[str, str]
    intents: Intents

    compress: APINullable[bool] = MISSING
    large_threshold: APINullable[int] = MISSING
    shard: APINullable[Tuple[int, int]] = MISSING
    presence: APINullable[...] = MISSING  # FIXME


@dataclass
class Resume(APIObject):
    """
    Used to replay missed events when a disconnected client resumes.

    :param token:
        session token

    :param session_id:
        session id

    :param seq:
        last sequence number received
    """
    token: str
    session_id: str
    seq: int


@dataclass
class RequestGuildMembers(APIObject):
    """
    Used to request all members for a guild or a list of guilds.

    :param guild_id:
        id of the guild to get members for

    :param query:
        string that username starts with, or an empty string
        to return all members

    :param limit:
        maximum number of members to send matching the `query`;
        a limit of `0` can be used with an empty string `query`
        to return all members

    :param presences:
        used to specify if we want the presences of the matches members

    :param user_ids:
        used to specify which users you wish to fetch

    :param nonce:
        nonce to identify the Guild Members Chunk response
    """
    guild_id: Snowflake
    limit: int

    query: APINullable[str] = MISSING
    presences: APINullable[bool] = MISSING
    user_ids: APINullable[Union[Snowflake, List[Snowflake]]] = MISSING
    nonce: APINullable[str] = MISSING


@dataclass
class UpdateVoiceState(APIObject):
    """
    Sent when a client wants to join, move,
    or disconnect from a voice channel.

    :param guild_id:
        id of the guild

    :param channel_id:
        id of the voice channel client
        wants to join (null if disconnecting)

    :param self_mute:
        is the client muted

    :param self_deaf:
        is the client deafened
    """
    guild_id: Snowflake
    self_mute: bool
    self_deaf: bool

    channel_id: Optional[Snowflake] = None


class StatusType(Enum):
    """
    :param online:
        Online

    :param dnd:
        Do Not Disturb

    :param idle:
        AFK

    :param invisible:
        Invisible and shown as offline

    :param offline:
        Offline
    """
    online = auto()
    dnd = auto()
    idle = auto()
    invisible = auto()
    offline = auto()


@dataclass
class UpdatePresence(APIObject):
    """
    Sent by the client to indicate a presence or status update.

    :param since:
        unix time (in milliseconds) of when the client went idle,
        or null if the client is not idle

    :param activities:
        the user's activities

    :param status:
        the user's new status

    :param afk:
        whether or not the client is afk
    """
    activities: List[Activity]
    status: StatusType
    afk: bool
    since: Optional[int] = None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
non-subscription event sent immediately after connecting,
contains server information
"""

from ..commands import ChatCommandHandler
from ..core.dispatch import GatewayDispatch
from ..objects import User
from ..utils import Coro


async def on_ready_middleware(self, payload: GatewayDispatch):
    """
    Middleware for ``on_ready`` event.

    :param self:
        The current client.

    :param payload:
        The data received from the ready event.
    """
    self.bot = User.from_dict(
        {"_client": self, "_http": self.http} | payload.data.get("user")
    )
    await ChatCommandHandler(self).initialize()
    return "on_ready",


def export() -> Coro:
    return on_ready_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.objects.invite import InviteTargetType
from pincer.objects.user import User
from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.timestamp import Timestamp
from pincer.utils.types import APINullable, MISSING


@dataclass
class InviteCreateEvent(APIObject):
    """
    Sent when a new invite to a channel is created.

    :param channel_id:
        the channel the invite is for

    :param code:
        the unique invite code

    :param created_at:
        the time at which the invite was created

    :param guild_id:
        the guild of the invite

    :param inviter:
        the user that created the invite

    :param max_age:
        how long the invite is valid for (in seconds)

    :param max_uses:
        the maximum number of times the invite can be used

    :param target_type:
        the type of target for this voice channel invite

    :param target_user:
        the user whose stream to display for
        this voice channel stream invite

    :param target_application:
        the embedded application to open for this
        voice channel embedded application invite

    :param temporary:
        whether or not the invite is temporary (invited users will
        be kicked on disconnect unless they're assigned a role)

    :param uses:
        how many times the invite has been used (always will be 0)
    """
    channel_id: Snowflake
    code: str
    created_at: Timestamp
    max_age: int
    max_uses: int
    temporary: bool

    guild_id: APINullable[Snowflake] = MISSING
    inviter: APINullable[User] = MISSING
    target_type: APINullable[InviteTargetType] = MISSING
    target_user: APINullable[User] = MISSING
    uses: int = 0


@dataclass
class InviteDeleteEvent(APIObject):
    """
    Sent when an invite is deleted.

    :param channel_id:
        the channel of the invite

    :param guild_id:
        the guild of the invite

    :param code:
        the unique invite code
    """
    channel_id: Snowflake
    code: str

    guild_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ..utils.api_object import APIObject


@dataclass
class VoiceRegion(APIObject):
    """
    Represents a Discord Voice Region object

    :param id:
        unique ID for the region

    :param name:
        name of the region

    :param vip:
        true if this is a vip-only server

    :param optimal:
        true for a single server
        that is closest to the current user's client

    :param deprecated:
        whether this is a deprecated voice region
        (avoid switching to these)

    :param custom:
        whether this is a custom voice region
        (used for events/etc)
    """
    id: str
    name: str
    vip: bool
    optimal: bool
    deprecated: bool
    custom: bool



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.objects.guild_member import GuildMember
from pincer.utils.api_object import APIObject
from pincer.utils.types import APINullable, MISSING
from pincer.utils.snowflake import Snowflake


@dataclass
class TypingStartEvent(APIObject):
    """
    Sent when a user starts typing in a channel.

    :param channel_id:
        id of the channel

    :param guild_id:
        id of the guild

    :param user_id:
        id of the user

    :param timestamp:
        unix time (in seconds) of when the user started typing

    :param member:
        the member who started typing if this happened in a guild
    """
    channel_id: Snowflake
    user_id: Snowflake
    timestamp: int

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import List

from .button import ButtonStyle
from .emoji import Emoji
from .select_menu import SelectOption
from ..utils import APIObject, MISSING, APINullable


@dataclass
class MessageComponent(APIObject):
    """
    Represents a Discord Message Component object

    :param type:
        component type

    :param custom_id:
        a developer-defined identifier for the component,
        max 100 characters

    :param disabled:
        whether the component is disabled,
        defaults to `False`

    :param style:
        one of button styles

    :param label:
        text that appears on the button, max 80 characters

    :param emoji:
        `name`, `id`, and `animated`

    :param url:
        a url for link-style buttons

    :param options:
        the choices in the select, max 25

    :param placeholder:
        custom placeholder text if nothing is selected,
        max 100 characters

    :param min_values:
        the minimum number of items that must be chosen;
        default 1, min 0, max 25

    :param max_values:
        the maximum number of items that can be chosen;
        default 1, max 25

    :param components:
        a list of child components
    """
    type: int
    options: List[SelectOption]

    custom_id: APINullable[str] = MISSING
    disabled: APINullable[bool] = False
    style: APINullable[ButtonStyle] = MISSING
    label: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    url: APINullable[str] = MISSING
    placeholder: APINullable[str] = MISSING
    min_values: APINullable[int] = 1
    max_values: APINullable[int] = 1
    components: APINullable[List[MessageComponent]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
non-subscription event sent when there is an error,
including command responses
"""
from ..core.dispatch import GatewayDispatch
from ..objects.events.error import DiscordError


def error_middleware(self, payload: GatewayDispatch):
    """
    Middleware for ``on_error`` event.

    :param client:

    :param payload:
        The data received from the ready event.
    """

    return "on_error",  [
        DiscordError.from_dict(
            {"_client": self, "_http": self.http} | payload.data
        )
    ]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import Union
from asyncio import sleep
from inspect import isasyncgenfunction
import logging

from ..commands import ChatCommandHandler
from ..exceptions import RateLimitError
from ..core.dispatch import GatewayDispatch
from ..objects import Interaction, Embed, Message, InteractionFlags
from ..utils import MISSING, should_pass_cls, Coro
from ..utils.extraction import get_params

_log = logging.getLogger(__name__)


async def interaction_create_middleware(self, payload: GatewayDispatch):
    """
    Middleware for ``on_interaction``, which handles command
    execution.

    :param self:
        The current client.

    :param payload:
        The data received from the interaction event.
    """

    interaction: Interaction = Interaction.from_dict(
        payload.data | {"_client": self, "_http": self.http}
    )

    command = ChatCommandHandler.register.get(interaction.data.name)

    if command:
        defaults = {param: None for param in get_params(command.call)}
        params = {}

        if interaction.data.options is not MISSING:
            params = {
                opt.name: opt.value for opt in interaction.data.options
            }

        kwargs = {**defaults, **params}

        if should_pass_cls(command.call):
            kwargs["self"] = self

        if isasyncgenfunction(command.call):
            message = command.call(**kwargs)
            started = False

            async for msg in message:

                msg = convert_message(self, msg)
                if started:
                    try:
                        await self.http.post(
                            f"webhooks/{interaction.application_id}"
                            f"/{interaction.token}",
                            msg.to_dict().get("data")
                        )
                    except RateLimitError as e:
                        _log.exception(
                            f"RateLimitError: {e}."
                            f" Retrying in {e.json.get('retry_after', 40)}"
                            f" seconds"
                        )

                        await sleep(e.json.get("retry_after", 40))
                        await self.http.post(
                            f"webhooks/{interaction.application_id}"
                            f"/{interaction.token}",
                            msg.to_dict().get("data")
                        )

                else:
                    started = True

                    await self.http.post(
                        f"interactions/{interaction.id}"
                        f"/{interaction.token}/callback",
                        msg.to_dict()
                    )
                await sleep(0.3)
        else:
            message = await command.call(**kwargs)
            message = convert_message(self, message)

            await self.http.post(
                f"interactions/{interaction.id}/{interaction.token}/callback",
                message.to_dict()
            )

    return "on_interaction_create", [interaction]


def convert_message(self, message: Union[Embed, Message, str]) -> Message:
    """Converts a message to a Message object"""
    if isinstance(message, Embed):
        message = Message(embeds=[message])
    elif not isinstance(message, Message):
        message = Message(message) if message else Message(
            self.received_message,
            flags=InteractionFlags.EPHEMERAL
        )
    return message


def export() -> Coro:
    return interaction_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class MessageReference(APIObject):
    """
    Represents a Discord Message Reference object

    :param message_id:
        id of the originating message

    :param channel_id:
        id of the originating message's channel

    :param guild_id:
        id of the originating message's guild

    :param fail_if_not_exists:
        when sending, whether to error if the referenced message doesn't
        exist instead of sending as a normal (non-reply) message,
        default true
    """
    message_id: APINullable[Snowflake] = MISSING
    channel_id: APINullable[Snowflake] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    fail_if_not_exists: APINullable[bool] = True



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, TYPE_CHECKING

from ..utils import APIObject, APINullable, MISSING, Snowflake, convert
from ..core.http import HTTPClient

if TYPE_CHECKING:
    from pincer import Client


class PremiumTypes(IntEnum):
    """
    The type of Discord premium a user has.
    """
    NONE = 0
    NITRO_CLASSIC = 1
    NITRO = 2


class VisibilityType(IntEnum):
    """
    The type of a connection visibility.
    """
    NONE = 0
    EVERYONE = 1


@dataclass
class User(APIObject):
    """
    Represents a Discord user. This can be a bot account or a
    human account.

    :param _client:
        reference to the Client

    :param _http:
        reference to the HTTPClient

    :param avatar:
        the user's avatar hash

    :param discriminator:
        the user's 4-digit discord-tag

    :param flags:
        the flags on a user's account

    :param id:
        the user's id

    :param username:
        the user's username, not unique across the platform

    :param accent_color:
        the user's banner color encoded as an integer representation of
        hexadecimal color code

    :param banner:
        the user's banner, or null if unset

    :param banner_color:
        the color of the user's banner

    :param bot:
        whether the user belongs to an OAuth2 application

    :param email:
        the user's email

    :param locale:
        the user's chosen language option

    :param mfa_enabled:
        whether the user has two factor enabled on their account

    :param premium_type:
        the type of Nitro subscription on a user's account

    :param public_flags:
        the public flags on a user's account

    :param system:
        whether the user is an Official Discord System user
        (part of the urgent message system)

    :param verified:
        whether the email on this account has been verified
    """

    _client: Client
    _http: HTTPClient

    avatar: Optional[str]
    discriminator: str
    id: Snowflake
    username: str

    flags: APINullable[int] = MISSING
    accent_color: APINullable[Optional[int]] = MISSING
    banner: APINullable[Optional[str]] = MISSING
    banner_color: APINullable[Optional[int]] = MISSING
    bot: APINullable[bool] = MISSING
    email: APINullable[Optional[str]] = MISSING
    locale: APINullable[str] = MISSING
    mfa_enabled: APINullable[bool] = MISSING
    premium_type: APINullable[int] = MISSING
    public_flags: APINullable[int] = MISSING
    system: APINullable[bool] = MISSING
    verified: APINullable[bool] = MISSING

    @property
    def premium(self) -> APINullable[PremiumTypes]:
        """
        The user their premium type in a usable enum.
        """
        return (
            MISSING
            if self.premium_type is MISSING
            else PremiumTypes(self.premium_type)
        )

    @property
    def mention(self) -> str:
        return f"<@!{self.id}>"

    def __str__(self):
        """Return the discord tag when object gets used as a string."""
        return self.username + '#' + self.discriminator

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)

    @classmethod
    async def from_id(cls, client: Client, _id: int) -> User:
        data = await client.http.get(f"users/{_id}")
        return cls.from_dict(data)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass

from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp


@dataclass
class ThreadMetadata(APIObject):
    """
    Represents a Discord Thread Metadata object

    :param archived:
        whether the thread is archived

    :param auto_archive_duration:
        duration in minutes to automatically archive the thread
        after recent activity, can be set to: 60, 1440, 4320, 10080

    :param archive_timestamp:
        timestamp when the thread's archive status was last changed,
        used for calculating recent activity

    :param locked:
        whether the thread is locked; when a thread is locked,
        only users with MANAGE_THREADS can unarchive it

    :param invitable:
        whether non-moderators can add other non-moderators to a thread;
        only available on private threads
    """
    archived: bool
    auto_archive_duration: int
    archive_timestamp: Timestamp
    locked: bool

    invitable: APINullable[bool] = MISSING


@dataclass
class ThreadMember(APIObject):
    """
    Represents a Discord Thread Member object

    :param join_timestamp:
        the time the current user last joined the thread

    :param flags:
        any user-thread settings, currently only used for notifications

    :param id:
        id of the thread

    :param user_id:
        id of the user
    """
    join_timestamp: Timestamp
    flags: int

    id: APINullable[Snowflake] = MISSING
    user_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when the user clicks a Rich Presence spectate invite in chat to
spectate a game
"""

# TODO: Implement event




# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp


class IntegrationExpireBehavior(IntEnum):
    """Represents a Discord Integration expire behavior"""
    REMOVE_ROLE = 0
    KICK = 1


@dataclass
class IntegrationAccount(APIObject):
    """
    Represents a Discord Integration Account object

    :param id:
        id of the account

    :param name:
        name of the account
    """
    id: str
    name: str


@dataclass
class IntegrationApplication(APIObject):
    """
    Represents a Discord Integration Application object

    :param id:
        the id of the app

    :param name:
        the name of the app

    :param icon:
        the icon hash of the app

    :param description:
        the description of the app

    :param summary:
        the summary of the app

    :param bot:
        the bot associated with this application
    """
    id: Snowflake
    name: str
    icon: Optional[str]
    description: str
    summary: str
    bot: APINullable[User] = MISSING


@dataclass
class Integration(APIObject):
    """
    Represents a Discord Integration object

    :param id:
        integration id

    :param name:
        integration name

    :param type:
        integration type (twitch, youtube, or discord)$

    :param enabled:
        is this integration enabled

    :param syncing:
        is this integration syncing

    :param role_id:
        id that this integration uses for subscribers

    :param enable_emoticons:
        whether emoticons should be synced for this integration
        (twitch only currently)

    :param expire_behavior:
        the behavior of expiring subscribers

    :param expire_grace_period:
        the grace period (in days) before expiring subscribers

    :param user:
        user for this integration

    :param account:
        integration account information

    :param synced_at:
        when this integration was last synced

    :param subscriber_count:
        how many subscribers this integration has

    :param revoked:
        has this integration been revoked

    :param application:
        The bot/OAuth2 application for discord integrations
    """

    id: Snowflake
    name: str
    type: str
    enabled: bool
    account: IntegrationAccount

    syncing: APINullable[bool] = MISSING
    role_id: APINullable[Snowflake] = MISSING
    enable_emoticons: APINullable[bool] = MISSING
    expire_behavior: APINullable[IntegrationExpireBehavior] = MISSING
    expire_grace_period: APINullable[int] = MISSING
    user: APINullable[User] = MISSING
    synced_at: APINullable[Timestamp] = MISSING
    subscriber_count: APINullable[int] = MISSING
    revoked: APINullable[bool] = MISSING
    application: APINullable[IntegrationApplication] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass, MISSING
from typing import List

from ..objects import Integration
from ..objects.user import VisibilityType
from ..utils import APIObject, APINullable


@dataclass
class Connection(APIObject):
    """
    The connection object that the user has attached.

    :param id:
        id of the connection account

    :param name:
        the username of the connection account

    :param type:
        the service of the connection (twitch, youtube)

    :param revoked:
        whether the connection is revoked

    :param integrations:
        an array of partial server integrations

    :param verified:
        whether the connection is verified

    :param friend_sync:
        whether friend sync is enabled for this connection

    :param show_activity:
        whether activities related to this connection
        will be shown in presence updates
    """
    id: str
    name: str
    type: str
    verified: bool
    friend_sync: bool
    show_activity: bool
    visibility: VisibilityType

    revoked: APINullable[bool] = MISSING
    integrations: APINullable[List[Integration]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .api_object import APIObject
from .conversion import convert
from .extraction import get_index, get_signature_and_params
from .insertion import should_pass_cls, should_pass_ctx
from .snowflake import Snowflake
from .timestamp import Timestamp
from .types import APINullable, Coro, MISSING


__all__ = (
    "APINullable", "APIObject", "convert", "Coro", "get_index",
    "get_signature_and_params", "MISSING", "should_pass_cls",
    "should_pass_ctx", "Snowflake", "Timestamp"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is created in a subscribed text channel"""

from ..core.dispatch import GatewayDispatch
from ..objects import UserMessage


async def message_create_middleware(self, payload: GatewayDispatch):
    return "on_message", [
        UserMessage.from_dict(
            {"_client": self, "_http": self.http} | payload.data
        )
    ]


def export():
    return message_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Any, List, Optional

from .presence import PresenceUpdateEvent
from ...objects.guild_member import GuildMember
from ...objects.emoji import Emoji
from ...objects.role import Role
from ...objects.sticker import Sticker
from ...objects.user import User
from ...utils.api_object import APIObject
from ...utils.snowflake import Snowflake
from ...utils.timestamp import Timestamp
from ...utils.types import MISSING, APINullable


@dataclass
class GuildBanAddEvent(APIObject):
    """
    Sent when a user is banned from a guild.

    :param guild_id:
        id of the guild

    :param user:
        the banned user
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildBanRemoveEvent(APIObject):
    """
    Sent when a user is unbanned from a guild.

    :param guild_id:
        id of the guild

    :param user:
        the unbanned user
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildEmojisUpdateEvent(APIObject):
    """
    Sent when a guild's emojis have been updated.

    :param guild_id:
        id of the guild

    :param emojis:
        array of emojis
    """
    guild_id: Snowflake
    emojis: List[Emoji]


@dataclass
class GuildStickersUpdateEvent(APIObject):
    """
    Sent when a guild's stickers have been updated.

    :param guild_id:
        id of the guild

    :param stickers:
        array of stickers
    """
    guild_id: Snowflake
    stickers: List[Sticker]


@dataclass
class GuildIntegrationsUpdateEvent(APIObject):
    """
    Sent when a guild integration is updated.

    :param guild_id:
        id of the guild whose integrations were updated
    """
    guild_id: Snowflake


@dataclass
class GuildMemberRemoveEvent(APIObject):
    """
    Sent when a user is removed from a guild (leave/kick/ban).

    :param guild_id:
        the id of the guild

    :param user:
        the user who was removed
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildMemberUpdateEvent(APIObject):
    """
    Sent when a guild member is updated.
    This will also fire when the user object
    of a guild member changes.

    :param guild_id:
        the id of the guild

    :param roles:
        user role ids

    :param user:
        the user

    :param nick:
        nickname of the user in the guild

    :param joined_at:
        when the user joined the guild

    :param premium_since:
        when the user started boosting the guild

    :param deaf:
        whether the user is deafened in voice channels

    :param mute:
        whether the user is muted in voice channels

    :param pending:
        whether the user has not yet passed the guild's
        Membership Screening requirements
    """
    guild_id: Snowflake
    roles: List[Snowflake]
    user: User
    nick: APINullable[Optional[str]] = MISSING
    joined_at: Optional[Timestamp] = None
    premium_since: APINullable[Optional[Timestamp]] = MISSING
    deaf: APINullable[bool] = MISSING
    mute: APINullable[bool] = MISSING
    pending: APINullable[bool] = MISSING


@dataclass
class GuildMembersChunkEvent(APIObject):
    """
    Sent in response to Guild Request Members.
    You can use the `chunk_index` and `chunk_count`
    to calculate how many chunks are left for your request.

    :param guild_id:
        the id of the guild

    :param members:
        set of guild members

    :param chunk_index:
        the chunk index in the expected chunks for this response
        (0 <= chunk_index < chunk_count)

    :param chunk_count:
        the total number of expected chunks for this response

    :param not_found:
        if passing an invalid id to `REQUEST_GUILD_MEMBERS`,
        it will be returned here

    :param presences:
        if passing true to `REQUEST_GUILD_MEMBERS`, presences
        of the returned members will be here

    :param nonce:"
        the nonce used in the Guild Members Request
    """
    guild_id: Snowflake
    members: List[GuildMember]
    chunk_index: int
    chunk_count: int

    not_found: APINullable[List[Any]] = MISSING
    presences: APINullable[PresenceUpdateEvent] = MISSING
    nonce: APINullable[str] = MISSING


@dataclass
class GuildRoleCreateEvent(APIObject):
    """
    Sent when a guild role is created.

    :param guild_id:
        the id of the guild

    :param role:
        the role created
    """
    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleUpdateEvent(APIObject):
    """
    Sent when a guild role is updated.

    :param guild_id:
        the id of the guild

    :param role:
        the role updated
    """
    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleDeleteEvent(APIObject):
    """
    Sent when a guild role is deleted.

    :param guild_id:
        id of the guild

    :param role_id:
        id of the role
    """
    guild_id: Snowflake
    role_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake


@dataclass
class WebhookUpdateEvent(APIObject):
    """
    Sent when a guild's channel webhook
    is created, updated, or deleted.

    :param guild_id:
        id of the guild

    :param channel_id:
        id of the channel
    """
    guild_id: Snowflake
    channel_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from ..objects.emoji import Emoji
from ..utils import APIObject, APINullable, MISSING


class ButtonStyle(IntEnum):
    """
    Buttons come in a variety of styles to convey different types of actions.
    These styles also define what fields are valid for a button.

    Primary:
        - color: blurple
        - required_field: custom_id
    Secondary:
        - color: gray
        - required_field: custom_id
    Success:
        - color: green
        - required_field: custom_id
    Danger:
        - color: red
        - required_field: custom_id
    Link:
        - color: gray, navigates to a URL
        - required_field: url
    """
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


@dataclass
class Button(APIObject):
    """
    Represents a Discord Button object.
    Buttons are interactive components that render on messages.

    They can be clicked by users,
    and send an interaction to your app when clicked.

    :param type:
        `2` for a button

    :param style:
        one of button styles

    :param label:
        text that appears on the button, max 80 characters

    :param emoji:
        `name`, `id`, and `animated`

    :param custom_id:
        a developer-defined identifier for the button,
        max 100 characters

    :param url:
        a url for link-style buttons

    :param disabled:
        whether the button is disabled (default `False`)
    """
    type: int
    style: ButtonStyle

    label: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    custom_id: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    disabled: APINullable[bool] = False



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import List

from pincer.objects.emoji import Emoji
from pincer.objects.guild_member import GuildMember
from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.types import APINullable, MISSING


@dataclass
class MessageDeleteEvent(APIObject):
    """
    Sent when a message is deleted.

    :param id:
        the id of the message

    :param channel_id:
        the id of the channel

    :param guild_id:
        the id of the guild
    """
    id: Snowflake
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageDeleteBulkEvent(APIObject):
    """
    Sent when multiple messages are deleted at once.

    :param ids:
        the ids of the messages

    :param channel_id:
        the id of the channel

    :param guild_id:
        the id of the guild
    """
    ids: List[Snowflake]
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionAddEvent(APIObject):
    """
    Sent when a user adds a reaction to a message.

    :param user_id:
        the id of the user

    :param channel_id:
        the id of the channel

    :param message_id:
        the id of the message

    :param guild_id:
        the id of the guild

    :param member:
        the member who reacted if this happened in a guild

    :param emoji:
        the emoji used to react
    """
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING


@dataclass
class MessageReactionRemoveEvent(APIObject):
    """
    Sent when a user removes a reaction from a message.

    :param user_id:
        the id of the user

    :param channel_id:
        the id of the channel

    :param message_id:
        the id of the message

    :param guild_id:
        the id of the guild

    :param emoji:
        the emoji used to react
    """
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionRemoveAllEvent(APIObject):
    """
    Sent when a user explicitly removes all reactions from a message.

    :param channel_id:
        the id of the channel

    :param message_id:
        the id of the message

    :param guild_id:
        the id of the guild
    """
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionRemoveEmojiEvent(APIObject):
    """
    Sent when a bot removes all instances of a given
    emoji from the reactions of a message.

    :param channel_id:
        the id of the channel

    :param guild_id:
        the id of the guild

    :param message_id:
        the id of the message

    :param emoji:
        the emoji that was removed
    """
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from .channel import Channel
from .guild import Guild
from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake


class WebhookType(IntEnum):
    """
    Represents the type of a webhook.

    :param INCOMING:
        Incoming Webhooks can post messages to channels with a
        generated token.

    :param CHANNEL_FOLLOWER:
        Channel Follower Webhooks are internal webhooks used with
        Channel Following to post new messages into channels.

    :param APPLICATION:
        Application webhooks are webhooks used with Interactions
    """
    INCOMING = 1
    CHANNEL_FOLLOWER = 2
    APPLICATION = 3


@dataclass
class Webhook(APIObject):
    """
    Represents a Discord channel webhook.

    :param id:
        the id of the webhook

    :param type:
        the type of the webhook

    :param channel_id:
        the channel id this webhook is for, if any

    :param name:
        the default name of the webhook

    :param avatar:
        the default user avatar hash of the webhook

    :param application_id:
        the bot/OAuth2 application that created this webhook

    :param user:
        the user this webhook was created by
        (not returned when getting a webhook with its token)

    :param token:
        the secure token of the webhook
        (returned for Incoming Webhooks)

    :param source_guild:
        the guild of the channel that this webhook is following
        (returned for Channel Follower Webhooks)

    :param source_channel:
        the channel that this webhook is following
        (returned for Channel Follower Webhooks)

    :param url:
        the url used for executing the webhook
        (returned by the webhooks OAuth2 flow)

    :param guild_id:
        the guild id this webhook is for, if any
    """

    id: Snowflake
    type: WebhookType

    channel_id: Optional[Snowflake] = None
    name: Optional[str] = None
    avatar: Optional[str] = None
    application_id: Optional[Snowflake] = None

    user: APINullable[User] = MISSING
    token: APINullable[str] = MISSING
    source_guild: APINullable[Guild] = MISSING
    source_channel: APINullable[Channel] = MISSING
    url: APINullable[str] = MISSING

    guild_id: APINullable[Optional[Snowflake]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.timestamp import Timestamp
from pincer.utils.types import MISSING, APINullable


@dataclass
class ChannelPinsUpdateEvent(APIObject):
    """
    Sent when a message is pinned or unpinned in a text channel.
    This is not sent when a pinned message is deleted.

    :param guild_id:
        the id of the guild

    :param channel_id:
        the id of the channel

    :param last_pin_timestamp:
        the time at which the most recent pinned message was pinned
    """
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING
    last_pin_timestamp: APINullable[Timestamp] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class Attachment(APIObject):
    """
    Represents a Discord Attachment object

    :param id:
        attachment id

    :param filename:
        name of file attached

    :param content_type:
        the attachment's data type

    :param size:
        size of file in bytes

    :param url:
        source url of file

    :param proxy_url:
        a proxied url of file

    :param height:
        height of file (if image)

    :param width:
        width of file (if image)
    """
    id: Snowflake
    filename: str
    size: int
    url: str
    proxy_url: str

    content_type: APINullable[str] = MISSING
    height: APINullable[Optional[int]] = MISSING
    width: APINullable[Optional[int]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is deleted in a subscribed text channel"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.message import MessageDeleteEvent


async def on_message_delete_middleware(self, payload: GatewayDispatch):
    """
    Middleware for ``on_message_delete`` event.

    :param self:
        The current client.

    :param payload:
        The data received from the delete message event.
    """
    return "on_message_delete", [
        MessageDeleteEvent.from_dict(
            {"_client": self, "_http": self.http} | payload.data
        )
    ]


def export():
    return on_message_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from enum import Enum, auto


class GuildFeatures(Enum):
    """Represents Guild Features strings."""
    ANIMATED_ICON = auto()
    BANNER = auto()
    COMMERCE = auto()
    COMMUNITY = auto()
    DISCOVERABLE = auto()
    FEATURABLE = auto()
    INVITE_SPLASH = auto()
    MEMBER_VERIFICATION_GATE_ENABLED = auto()
    NEWS = auto()
    PARTNERED = auto()
    PREVIEW_ENABLED = auto()
    VANITY_URL = auto()
    VERIFIED = auto()
    VIP_REGIONS = auto()
    WELCOME_SCREEN_ENABLED = auto()
    TICKETED_EVENTS_ENABLED = auto()
    MONETIZATION_ENABLED = auto()
    MORE_STICKERS = auto()
    THREE_DAY_THREAD_ARCHIVE = auto()
    SEVEN_DAY_THREAD_ARCHIVE = auto()
    PRIVATE_THREADS = auto()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List

from .role import Role
from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class Emoji(APIObject):
    """
    :param id:
        emoji id

    :param name:
        emoji name

    :param animated:
        whether this emoji is animated

    :param available:
        whether this emoji can be used, may be false due to loss of Server
        Boosts

    :param managed:
        whether this emoji is managed

    :param require_colons:
        whether this emoji must be wrapped in colons

    :param roles:
        roles allowed to use this emoji

    :param user:
        user that created this emoji
    """

    id: Optional[Snowflake]
    name: Optional[str]

    animated: APINullable[bool] = MISSING
    available: APINullable[bool] = MISSING
    managed: APINullable[bool] = MISSING
    require_colons: APINullable[bool] = MISSING
    roles: APINullable[List[Role]] = MISSING
    user: APINullable[User] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from ..utils import APIObject, Snowflake


class PrivacyLevel(IntEnum):
    """
    Represents the level of publicity of a stage.
    """
    PUBLIC = 1
    GUILD_ONLY = 2


@dataclass
class StageInstance(APIObject):
    """
    Represents a Stage Instance object


    :param id:
        id of this Stage instance

    :param guild_id:
        guild id of the associated Stage channel

    :param channel_id:
        id of the associated Stage channel

    :param topic:
        topic of the Stage instance (1-120 characters)

    :param privacy_level:
        privacy level of the Stage instance

    :param discoverable:
        is Stage Discovery enabled
    """
    id: Snowflake
    guild_id: Snowflake
    channel_id: Snowflake
    topic: str
    privacy_level: PrivacyLevel
    discoverable: bool

    @property
    def discoverable_disabled(self) -> bool:
        return not self.discoverable



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from .guild import Guild
from .user import User
from ..utils import APIObject, Snowflake, Timestamp


@dataclass
class GuildTemplate(APIObject):
    """
    Represents a code that when used,
    creates a guild based on a snapshot of an existing guild.

    :param code:
        the template code (unique ID)

    :param name:
        template name

    :param description:
        the description for the template

    :param usage_count:
        number of times this template has been used

    :param creator_id:
        the ID of the user who created the template

    :param creator: the
        user who created the template

    :param created_at:
        when this template was created

    :param updated_at:
        when this template was last synced to the source guild

    :param source_guild_id:
        the ID of the guild this template is based on

    :param serialized_source_guild:
        the guild snapshot this template contains

    :param is_dirty:
        whether the template has unsynced changes
    """
    code: str
    name: str
    description: Optional[str]
    usage_count: int
    creator_id: Snowflake
    creator: User
    created_at: Timestamp
    updated_at: Timestamp
    source_guild_id: Snowflake
    serialized_source_guild: Guild
    is_dirty: Optional[bool]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


from __future__ import annotations

from json import dumps, loads
from typing import Optional, Union, Dict, Any


class GatewayDispatch:
    """Represents a websocket message."""

    def __init__(
            self,
            op: int,
            data: Optional[Union[int, Dict[str, Any]]],
            seq: Optional[int] = None,
            name: Optional[str] = None
    ):
        """
        Instantiate a new GatewayDispatch object.

        :param op:
            The discord opcode which represents what the message means.

        :param data:
            The event data that has been sent/received.

        :param seq:
            The sequence number of a message, which can be used
            for resuming sessions and heartbeats.

        :param name:
            The event name for the payload.
        """
        self.op: int = op
        self.data: Optional[Union[int, Dict[str, Any]]] = data
        self.seq: Optional[int] = seq
        self.event_name: Optional[str] = name

    def __str__(self) -> str:
        """
        :return
            The string representation of the GatewayDispatch object.

        This can be used to send a websocket message to the gateway.
        """
        return dumps(
            dict(
                op=self.op,
                d=self.data,
                s=self.seq,
                t=self.event_name
            )
        )

    @classmethod
    def from_string(cls, payload: str) -> GatewayDispatch:
        """
        Parses a given payload from a string format
            and returns a GatewayDispatch.

        :param payload:
            The payload to parse.

        :return:
            A proper GatewayDispatch object.
        """
        payload: Dict[str, Union[int, str, Dict[str, Any]]] = loads(payload)
        return cls(
            payload.get("op"),
            payload.get("d"),
            payload.get("s"),
            payload.get("t")
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from ..utils import APIObject, Snowflake


@dataclass
class GuildWidget(APIObject):
    """
    Represents a Discord Guild Widget object

    :param enabled:
        whether the widget is enabled

    :param channel_id:
        the widget channel id
    """
    enabled: bool
    channel_id: Optional[Snowflake]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass

from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class RoleTags(APIObject):
    """
    Special tags/flags which have been defined for a role.

    :bot_id:
        The id of the bot this role belongs to.
        (the role got created by adding the bot with this id)

    :integration_id:
        The id of the integration this role belongs to.
        (the role got created by adding an integration with this id)

    :premium_subscriber:
        Whether this is the guild's premium subscriber role or not.
    """
    bot_id: APINullable[Snowflake] = MISSING
    integration_id: APINullable[Snowflake] = MISSING
    premium_subscriber: APINullable[bool] = MISSING


@dataclass
class Role(APIObject):
    """
    Represents a Discord guild/server role.

    :param color:
        integer representation of hexadecimal color code

    :param hoist:
        if this role is pinned in the user listing

    :param id:
        role id

    :param managed:
        whether this role is managed by an integration

    :param mentionable:
        whether this role is mentionable

    :param name:
        role name

    :param permissions:
        permission bit set

    :param position:
        position of this role

    :param tags:
        the tags this role has
    """

    color: int
    hoist: bool
    id: Snowflake
    managed: bool
    mentionable: bool
    name: str
    permissions: str
    position: int

    tags: APINullable[RoleTags] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from enum import IntEnum
from typing import Dict

from .user import User
from ..utils import APIObject, Snowflake, APINullable, MISSING


class CallbackType(IntEnum):
    """
    The types of response a client can give to a interaction.

    :param PONG:
        ACK a Ping

    :param MESSAGE:
        respond to an interaction with a message

    :param DEFERRED_MESSAGE:
        ACK an interaction and edit a response later,
        the user sees a loading state

    :param DEFERRED_UPDATE_MESSAGE:
        for components, ACK an interaction
        and edit the original message later;
        the user does not see a loading state

    :param UPDATE_MESSAGE
        for components, edit the message the component was attached to
    """
    PONG = 1
    MESSAGE = 4
    DEFERRED_MESSAGE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7


class InteractionType(IntEnum):
    """
    Represents the different types of interactions the client
    can have with a member.

    :param CHAT_INPUT:
        Slash commands;
        a text-based command that shows up when a user types /

    :param USER:
        A UI-based command that shows up when you right click
        or tap on a user

    :param MESSAGE:
        A UI-based command that shows up when you right click
        or tap on a message
    """
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3


@dataclass
class MessageInteraction(APIObject):
    """
    Represents a Discord Message Interaction object

    This is sent on the message object when the message
    is a response to an Interaction without an existing message.

    :param id: id of the interaction
    :param type: the type of interaction
    :param name: the name of the application command
    :param user: the user who invoked the interaction
    """
    id: Snowflake
    type: InteractionType
    name: str
    user: User

    member: APINullable[Dict] = MISSING
